# (c) M_Chabot 2016 ...
# Permission to use, copy, modify, and distribute this software for
# any purpose and without fee is hereby granted, provided the above
# copyright notice appear in all copies.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
# ____________________________________________________________________________
# Description
# This stems from TIG's 'dropverts.rb' where all the logic and coding has been set; moving endpoints in the Z axis.
# I just ported it to add X and Y axies as a complement.
# and packaged the 3 together as a tool set. Naming it, from a user point of view, 'Align Endpoints'.
# added 3 most frequents (imho) to context menu.
# added LanguageHandler and french strings.
# ____________________________________________________________________________
=begin
Copyright 2011-2015, TIG (c)
Permission to use, copy, modify, and distribute this software for any
purpose, and currently without fee, is hereby granted, provided that
this text and the above copyright (c) notice appear in all copies.
This software is provided "as is" and without any express or implied
warranties, including, without limitation, the implied warranties of
merchantability and fitness for a particular purpose.
###
Name: dropverts.rb
###
Menu: 'Plugins' > 'Drop Vertices...' >  changed to 'Align Endpoints' - mc.
'To Nearest Object'
'To Nearest Below'
'To Nearest Above'
'To Nearest+Z=?'
'To Lowest'
'To Highest'
'To Middle'
'To Average'
'To Z=?'
'To Z=%'
###
Usage:
Preselect the edges where you want vertices to 'adjust', choose tool...

'To Nearest Object' means all of the edges' vertices will 'drop' on to
the 'objects' below them, OR 'lift' to the 'objects' above them if there
was nothing below, or to Z=0 if there is nothing above/below.

'To Nearest Below' means all of the edges' vertices will 'drop' on to
the next 'object' below them, or to Z=0 if there is nothing below.

'To Nearest Above' means all of the edges' vertices will 'lift' up to
the next 'object' above them, or to Z=0 if there is nothing above.

'To Nearest+Z=?' means you enter the default drop Z value in a dialog
and it then does the equivalent of 'To Nearest Object' BUT stops at
the specified Z value [rather than Z=0] is nothing is above/below.

'To Lowest' means all of the edges' vertices will 'drop' to match the
Z value of the lowest selected vertex.

'To Highest' means all of the edges' vertices will 'lift' to match the
Z value of the highest selected vertex.

'To Middle' means all of the edges' vertices will 'adjust' to match the
mid Z value between the highest and lowest selected vertices.

'To Average' means all of the edges' vertices will 'adjust' to match the
average Z value of all of the selected vertices; this is not necessarily
the same as 'To Middle' as a disproportionate number of low vertices
in the selection will reduce the average but not affect the mid value.

'To Z=?' means a dialog asks for the Z value; enter a number to use
current-units e.g. 123 might be 123" or 123mm depending on the model's
current units; to use something other than the current units suffix the
number thus '123mm' to use 123mm when your units might be in 'meters':
the vertices will all adjust to that specified Z value: default Z=0.

'To Z=%' means that a dialog asks for the % [Percentage]; enter a
value: the default is 100 which means 'no change': entering a smaller
value, say 50, will adjust each selected vertex to be 50% its height
above the lowest selected vertex [i.e. the mesh is 'deflated']; entering
a higher value, say 150, will adjust each selected vertex to be 150%
its height above the lowest selected vertex [i.e. the mesh is
'inflated']; entering 0 will completely 'flatten' the mesh to the level
of the lowest selected vertex [equivalent to using 'To Lowest'];
entering negative values will 'flip' the mesh - so -100 will invert it,
-50 will invert it and deflate it, and -150 will invert it and inflate
it.
###
Donations: PayPal.com to info @ revitrev.org
###
Version:
1.0 20110806 First issue.
1.1 20110807 Nearest-Below/-Above, Middle/Average and Z=% options added.
1.2 20110807 Nearest+Z=? option added.
1.3 20110810 Glitch with Lowest/Highest/Middle fixed.
2.0 20150422 Glitches with dropping curves resolved.
=end

module McF3D ## changed the module, by fear of interfering in TIGs space.
require 'sketchup.rb'
require 'langhandler.rb'
###
def self.dropverts(z=nil,dn=nil,up=nil,pe=nil)
  ###
  z=z.to_f if z ###to units
  zN=true
  if ! up && ! dn
    zN=false
    up=true
	dn=true
  end
  ###
  model=Sketchup.active_model
  ents=model.active_entities
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e|
    e.explode_curve if e.curve
	verts << e.vertices
  }
  ss.clear
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  verts.flatten!
  verts.uniq!
  if pe ### percentage
    pe=1.0-(pe.to_f/100.0)
    z=verts[0].position.z
    verts.each{|v|z=v.position.z if v.position.z < z}
  end
  vzs=[]
  vts=[]
  verts.each{|v|
    vzs << v
    pv=v.position
    if z
      if ! pe
        if zN ### z passed but also 'to nearest'
          rayt=model.raytest(pv, Z_AXIS.reverse)
          if rayt
            pt=rayt[0]
          else
            rayt=model.raytest(pv, Z_AXIS)
            if rayt
              pt=rayt[0]
            else
              pt=pv.clone
              pt.z=z if z
              pt.z=0 if ! z
            end
          end
        else ### there's no zN - so z=z
          pt=pv.clone
          pt.z=z if z
          pt.z=0 if ! z
        end
      else ### we'll apply the percentage
        pt=pv.offset(Z_AXIS.reverse, (pv.z - z)*pe)
      end
    elsif dn
      rayt=model.raytest(pv, Z_AXIS.reverse)
      if rayt
        pt=rayt[0]
      elsif up
        rayt=model.raytest(pv, Z_AXIS)
        if rayt
          pt=rayt[0]
        else
          pt=pv.clone
          pt.z=z if z
          pt.z=0 if ! z
        end
      else
        pt=pv.clone
        pt.z=z if z
        pt.z=0 if ! z
      end
    elsif up
      rayt=model.raytest(pv, Z_AXIS)
      if rayt
        pt=rayt[0]
      else
        pt=pv.clone
        pt.z=z if z
        pt.z=0 if ! z
      end
    else###no z
      pt=pv.clone
      pt.z=z if z ### unlikely!
      pt.z=0 if ! z
    end#if
    vts << pv.vector_to(pt)
  }
  begin
    model.start_operation("McF3D.dropverts", true)
  rescue
    model.start_operation("McF3D.dropverts")
  end
    ents.transform_by_vectors(vzs,vts)
    return vzs ### returns array of vertices processed
  model.commit_operation
end
###
def self.dropverts_lowest()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  verts.flatten!
  verts.uniq!
  z=verts[0].position.z
  verts.each{|v|z=v.position.z if v.position.z < z}
  McF3D.dropverts(z,nil,nil,nil)
end
###
def self.dropverts_highest()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  verts.flatten!
  verts.uniq!
  z=verts[0].position.z
  verts.each{|v|z=v.position.z if v.position.z > z}
  McF3D.dropverts(z,nil,nil,nil)
end
###
def self.dropverts_middle()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  verts.flatten!
  verts.uniq!
  zz=verts[0].position.z
  zzz=verts[0].position.z
  verts.each{|v|
     zz=v.position.z if v.position.z > zz
    zzz=v.position.z if v.position.z < zzz
  }
  z=(zz-((zz-zzz)/2.0)).to_s+'"'
  McF3D.dropverts(z,nil,nil,nil)
end
###
def self.dropverts_average()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  verts.flatten!
  verts.uniq!
  zz=0.0
  verts.each{|v|zz = zz+v.position.z}
  z = (zz/(verts.length.to_f)).to_s+'"'
  McF3D.dropverts(z,nil,nil,nil)
end
###
def self.dropverts_dialog()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  @dropvertsZ="0" unless @dropvertsZ
  results=UI.inputbox([AeLH["To Z=?"]],[@dropvertsZ],AeLH["Drop Vertices ?"])
  return nil unless results
  @dropvertsZ=results[0]
  McF3D.dropverts(@dropvertsZ,nil,nil,nil)
end
###
def self.dropverts_nearestZ()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  @dropvertsZ="0" unless @dropvertsZ
  results=UI.inputbox([AeLH["To Z=?"]],[@dropvertsZ],AeLH["Drop Vertices [Nearest+]"])
  return nil unless results
  @dropvertsZ=results[0]
  McF3D.dropverts(@dropvertsZ,true,true,nil)
end
###
def self.dropverts_percentage()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  @dropvertsP=100.0 unless @dropvertsP
  results=UI.inputbox([AeLH["To Z=%"]],[@dropvertsP],AeLH["Drop Vertices [%]"])
  return nil unless results
  @dropvertsP=results[0]
  McF3D.dropverts(nil,nil,nil,@dropvertsP)
end
### ZZZZZZ done

### XXXXXX start
def self.alXverts(x=nil,dn=nil,up=nil,pe=nil)
  ###
  x=x.to_f if x ###to units
  xN=true
  if ! up && ! dn
    xN=false
    up=true
	dn=true
  end
  ###
  model=Sketchup.active_model
  ents=model.active_entities
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e|
    e.explode_curve if e.curve
	verts << e.vertices
  }
  ss.clear
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  verts.flatten!
  verts.uniq!
  if pe ### percentage
    pe=1.0-(pe.to_f/100.0)
    x=verts[0].position.x
    verts.each{|v|x=v.position.x if v.position.x < x}
  end
  vxs=[]
  vts=[]
  verts.each{|v|
    vxs << v
    pv=v.position
    if x
      if ! pe
        if xN ### x passed but also 'to nearest'
          rayt=model.raytest(pv, X_AXIS.reverse)
          if rayt
            pt=rayt[0]
          else
            rayt=model.raytest(pv, X_AXIS)
            if rayt
              pt=rayt[0]
            else
              pt=pv.clone
              pt.x=x if x
              pt.x=0 if ! x
            end
          end
        else ### there's no xN - so x=x
          pt=pv.clone
          pt.x=x if x
          pt.x=0 if ! x
        end
      else ### we'll apply the percentage
        pt=pv.offset(X_AXIS.reverse, (pv.x - x)*pe)
      end
    elsif dn
      rayt=model.raytest(pv, X_AXIS.reverse)
      if rayt
        pt=rayt[0]
      elsif up
        rayt=model.raytest(pv, X_AXIS)
        if rayt
          pt=rayt[0]
        else
          pt=pv.clone
          pt.x=x if x
          pt.x=0 if ! x
        end
      else
        pt=pv.clone
        pt.x=x if x
        pt.x=0 if ! x
      end
    elsif up
      rayt=model.raytest(pv, X_AXIS)
      if rayt
        pt=rayt[0]
      else
        pt=pv.clone
        pt.x=x if x
        pt.x=0 if ! x
      end
    else###no x
      pt=pv.clone
      pt.x=x if x ### unlikely!
      pt.x=0 if ! x
    end#if
    vts << pv.vector_to(pt)
  }
  begin
    model.start_operation("McF3D.alXverts", true)
  rescue
    model.start_operation("McF3D.alXverts")
  end
    ents.transform_by_vectors(vxs,vts)
    return vxs ### returns array of vertices processed
  model.commit_operation
end
###
def self.alXverts_lowest()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  verts.flatten!
  verts.uniq!
  x=verts[0].position.x
  verts.each{|v|x=v.position.x if v.position.x < x}
  McF3D.alXverts(x,nil,nil,nil)
end
###
def self.alXverts_highest()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  verts.flatten!
  verts.uniq!
  x=verts[0].position.x
  verts.each{|v|x=v.position.x if v.position.x > x}
  McF3D.alXverts(x,nil,nil,nil)
end
###
def self.alXverts_middle()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  verts.flatten!
  verts.uniq!
  xx=verts[0].position.x
  xxx=verts[0].position.x
  verts.each{|v|
     xx=v.position.x if v.position.x > xx
    xxx=v.position.x if v.position.x < xxx
  }
  x=(xx-((xx-xxx)/2.0)).to_s+'"'
  McF3D.alXverts(x,nil,nil,nil)
end
###
def self.alXverts_average()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  verts.flatten!
  verts.uniq!
  xx=0.0
  verts.each{|v|xx = xx+v.position.x}
  x = (xx/(verts.length.to_f)).to_s+'"'
  McF3D.alXverts(x,nil,nil,nil)
end
###
def self.alXverts_dialog()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  @alXvertsX="0" unless @alXvertsX
  results=UI.inputbox([AeLH["To X=?"]],[@alXvertsX],AeLH["Drop Vertices ?"])
  return nil unless results
  @alXvertsX=results[0]
  McF3D.alXverts(@alXvertsX,nil,nil,nil)
end
###
def self.alXverts_nearestX()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  @alXvertsX="0" unless @alXvertsX
  results=UI.inputbox([AeLH["To X=?"]],[@alXvertsX],AeLH["Drop Vertices [Nearest+]"])
  return nil unless results
  @alXvertsX=results[0]
  McF3D.alXverts(@alXvertsX,true,true,nil)
end
###
def self.alXverts_percentage()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  @alXvertsP=100.0 unless @alXvertsP
  results=UI.inputbox([AeLH["To X=%"]],[@alXvertsP], AeLH["Drop Vertices [%]"])
  return nil unless results
  @alXvertsP=results[0]
  McF3D.alXverts(nil,nil,nil,@alXvertsP)
end
### XXXXXX done

### YYYYYY start
def self.alYverts(y=nil,dn=nil,up=nil,pe=nil)
  ###
  y=y.to_f if y ###to units
  yN=true
  if ! up && ! dn
    yN=false
    up=true
	dn=true
  end
  ###
  model=Sketchup.active_model
  ents=model.active_entities
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e|
    e.explode_curve if e.curve
	verts << e.vertices
  }
  ss.clear
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  verts.flatten!
  verts.uniq!
  if pe ### percentage
    pe=1.0-(pe.to_f/100.0)
    y=verts[0].position.y
    verts.each{|v|y=v.position.y if v.position.y < y}
  end
  vys=[]
  vts=[]
  verts.each{|v|
    vys << v
    pv=v.position
    if y
      if ! pe
        if yN ### y passed but also 'to nearest'
          rayt=model.raytest(pv, Y_AXIS.reverse)
          if rayt
            pt=rayt[0]
          else
            rayt=model.raytest(pv, Y_AXIS)
            if rayt
              pt=rayt[0]
            else
              pt=pv.clone
              pt.y=y if y
              pt.y=0 if ! y
            end
          end
        else ### there's no yN - so y=y
          pt=pv.clone
          pt.y=y if y
          pt.y=0 if ! y
        end
      else ### we'll apply the percentage
        pt=pv.offset(Y_AXIS.reverse, (pv.y - y)*pe)
      end
    elsif dn
      rayt=model.raytest(pv, Y_AXIS.reverse)
      if rayt
        pt=rayt[0]
      elsif up
        rayt=model.raytest(pv, Y_AXIS)
        if rayt
          pt=rayt[0]
        else
          pt=pv.clone
          pt.y=y if y
          pt.y=0 if ! y
        end
      else
        pt=pv.clone
        pt.y=y if y
        pt.y=0 if ! y
      end
    elsif up
      rayt=model.raytest(pv, Y_AXIS)
      if rayt
        pt=rayt[0]
      else
        pt=pv.clone
        pt.y=y if y
        pt.y=0 if ! y
      end
    else###no y
      pt=pv.clone
      pt.y=y if y ### unlikely!
      pt.y=0 if ! y
    end#if
    vts << pv.vector_to(pt)
  }
  begin
    model.start_operation("McF3D.alYverts", true)
  rescue
    model.start_operation("McF3D.alYverts")
  end
    ents.transform_by_vectors(vys,vts)
    return vys ### returns array of vertices processed
  model.commit_operation
end
###
def self.alYverts_lowest()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  verts.flatten!
  verts.uniq!
  y=verts[0].position.y
  verts.each{|v|y=v.position.y if v.position.y < y}
  McF3D.alYverts(y,nil,nil,nil)
end
###
def self.alYverts_highest()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  verts.flatten!
  verts.uniq!
  y=verts[0].position.y
  verts.each{|v|y=v.position.y if v.position.y > y}
  McF3D.alYverts(y,nil,nil,nil)
end
###
def self.alYverts_middle()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  verts.flatten!
  verts.uniq!
  yy=verts[0].position.y
  yyy=verts[0].position.y
  verts.each{|v|
     yy=v.position.y if v.position.y > yy
    yyy=v.position.y if v.position.y < yyy
  }
  y=(yy-((yy-yyy)/2.0)).to_s+'"'
  McF3D.alYverts(y,nil,nil,nil)
end
###
def self.alYverts_average()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  verts.flatten!
  verts.uniq!
  yy=0.0
  verts.each{|v|yy = yy+v.position.y}
  y = (yy/(verts.length.to_f)).to_s+'"'
  McF3D.alYverts(y,nil,nil,nil)
end
###
def self.alYverts_dialog()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  @alYvertsY="0" unless @alYvertsY
  results=UI.inputbox([AeLH["To Y=?"]],[@alYvertsY],AeLH["Drop Vertices ?"])
  return nil unless results
  @alYvertsY=results[0]
  McF3D.alYverts(@alYvertsY,nil,nil,nil)
end
###
def self.alYverts_nearestY()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  @alYvertsY="0" unless @alYvertsY
  results=UI.inputbox([AeLH["To Y=?"]],[@alYvertsY],AeLH["Drop Vertices [Nearest+]"])
  return nil unless results
  @alYvertsY=results[0]
  McF3D.alYverts(@alYvertsY,true,true,nil)
end
###
def self.alYverts_percentage()
  model=Sketchup.active_model
  ss=model.selection
  verts=[]
  ss.grep(Sketchup::Edge).each{|e| verts << e.vertices }
  unless verts[0]
    UI.messagebox(AeLH["Pre-Select Edges!"])
    return nil
  end
  @alYvertsP=100.0 unless @alYvertsP
  results=UI.inputbox([AeLH["To Y=%"]],[@alYvertsP],AeLH["Drop Vertices [%]"])
  return nil unless results
  @alYvertsP=results[0]
  McF3D.alYverts(nil,nil,nil,@alYvertsP)
end
### YYYYYYY done

### menu
 unless file_loaded?(__FILE__)
  aemenu=UI.menu('Plugins').add_submenu(AeLH['Align Endpoints...'])
  ### menu X
  aex=aemenu.add_submenu(AeLH['Align X...'])
  aex.add_item(AeLH["To Nearest Object"]){McF3D.alXverts(nil,true,true,nil)}
  aex.add_item(AeLH["To Nearest Below"]){McF3D.alXverts(nil,true,nil,nil)}
  aex.add_item(AeLH["To Nearest Above"]){McF3D.alXverts(nil,nil,true,nil)}
  aex.add_item(AeLH["To Nearest+X=?"]){McF3D.alXverts_nearestX()}
  aex.add_item(AeLH["To Lowest"]){McF3D.alXverts_lowest()}
  aex.add_item(AeLH["To Highest"]){McF3D.alXverts_highest()}
  aex.add_item(AeLH["To Middle"]){McF3D.alXverts_middle()}
  aex.add_item(AeLH["To Average"]){McF3D.alXverts_average()}
  aex.add_item(AeLH["To X=?"]){McF3D.alXverts_dialog()}
  aex.add_item(AeLH["To X=%"]){McF3D.alXverts_percentage()}
  ### menu Y
  aey=aemenu.add_submenu(AeLH['Align Y...'])
  aey.add_item(AeLH["To Nearest Object"]){McF3D.alYverts(nil,true,true,nil)}
  aey.add_item(AeLH["To Nearest Below"]){McF3D.alYverts(nil,true,nil,nil)}
  aey.add_item(AeLH["To Nearest Above"]){McF3D.alYverts(nil,nil,true,nil)}
  aey.add_item(AeLH["To Nearest+Y=?"]){McF3D.alYverts_nearestY()}
  aey.add_item(AeLH["To Lowest"]){McF3D.alYverts_lowest()}
  aey.add_item(AeLH["To Highest"]){McF3D.alYverts_highest()}
  aey.add_item(AeLH["To Middle"]){McF3D.alYverts_middle()}
  aey.add_item(AeLH["To Average"]){McF3D.alYverts_average()}
  aey.add_item(AeLH["To Y=?"]){McF3D.alYverts_dialog()}
  aey.add_item(AeLH["To Y=%"]){McF3D.alYverts_percentage()}
  ### menu Z
  aez=aemenu.add_submenu(AeLH['Align Z...'])
  aez.add_item(AeLH["To Nearest Object"]){McF3D.dropverts(nil,true,true,nil)}
  aez.add_item(AeLH["To Nearest Below"]){McF3D.dropverts(nil,true,nil,nil)}
  aez.add_item(AeLH["To Nearest Above"]){McF3D.dropverts(nil,nil,true,nil)}
  aez.add_item(AeLH["To Nearest+Z=?"]){McF3D.dropverts_nearestZ()}
  aez.add_item(AeLH["To Lowest"]){McF3D.dropverts_lowest()}
  aez.add_item(AeLH["To Highest"]){McF3D.dropverts_highest()}
  aez.add_item(AeLH["To Middle"]){McF3D.dropverts_middle()}
  aez.add_item(AeLH["To Average"]){McF3D.dropverts_average()}
  aez.add_item(AeLH["To Z=?"]){McF3D.dropverts_dialog()}
  aez.add_item(AeLH["To Z=%"]){McF3D.dropverts_percentage()}
 file_loaded(__FILE__)

 ## add context menu for most frequent
  UI.add_context_menu_handler do|contmenu|
  contmenu.add_separator
  sub=contmenu.add_submenu(AeLH['Align Endpoints...'])
  sub.add_item(AeLH['To Nearest X']){McF3D.alXverts(nil,true,true,nil)}
  sub.add_item(AeLH['To Nearest Y']){McF3D.alYverts(nil,true,true,nil)}
  sub.add_item(AeLH['To Nearest Z']){McF3D.dropverts(nil,true,true,nil)}
  end
 end
end#module
